import os
from jinja2 import Environment
import networkx as nx


EXTS = [".jinja2", ".j2", ".sql"]


def is_ty(ty: str):
    def is_script(tname: str) -> bool:
        ext = os.path.splitext(tname)[1]
        if ext not in EXTS:
            return False
        if tname.startswith(f"{ty}/"):
            return True
        return False

    return is_script


def load_all(env: Environment, dag: nx.DiGraph, config, ty: str, field: str):
    scripts = {}

    first_connection = list(config["connections"].keys())[0]

    for tname in env.list_templates(filter_func=is_ty(field)):
        name = os.path.basename(tname)[:-4]
        scripts[name] = {
            "name": name,
            "type": ty,
            "template": tname,
            "data": {},
            "output": f"{name}.sql",
            "connection": first_connection,
        }

    for name, overrides in config.get(field, {}).items():
        data = scripts.setdefault(name, {})
        data.update(overrides)
        data.setdefault("data", {})
        data.setdefault("type", ty)
        data.setdefault("output", f"{name}.sql")
        data.setdefault("connection", first_connection)
        scripts[name] = data

    for name, data in scripts.items():
        dag.add_node(name, conn=data["connection"])

    return scripts


def load_all_scripts(env: Environment, dag, config):
    return load_all(env, dag, config, "script", "scripts")


def load_all_models(env: Environment, dag, config):
    return load_all(env, dag, config, "model", "models")



import os
import click
import yaml
from jinja2 import Environment, FileSystemLoader
import networkx as nx
from .models import load_all_models, load_all_scripts
from .macros import load_macros, load_macros_from_dir, load_macros_from_config
from .connections import build_connection
from dotenv import load_dotenv

env = Environment(loader=FileSystemLoader(["flow"]))


@click.group(chain=True, invoke_without_command=True)
@click.pass_context
def main(ctx):
    load_dotenv(".env")
    if ctx.invoked_subcommand is None:
        ctx.invoke(build)
        ctx.invoke(run)


@main.command("build")
def build():
    with open("build.yml") as fd:
        build_info = yaml.safe_load(fd)

    dag = nx.DiGraph()

    macros = load_macros(env, dag, build_info)
    env.globals.update(macros)
    macros = load_macros_from_dir(env)
    env.globals.update(macros)
    macros = load_macros_from_config(build_info)
    env.globals.update(macros)

    os.makedirs("build", exist_ok=True)

    models = {}
    models.update(load_all_models(env, dag, build_info))
    models.update(load_all_scripts(env, dag, build_info))

    for name, model in sorted(models.items()):
        template = env.get_template(model["template"])
        ty = model["type"]
        output = model["output"]
        fn = f"build/{output}"
        print(f"render {ty} {fn}")
        env.globals.update(current_node=name)
        rendered = template.render(model["data"])
        with open(fn, "w") as fd:
            fd.write(rendered)

    with open("build/graph.yml", "w") as fd:
        yaml.safe_dump(
            {
                "edges": list(dag.edges()),
                "nodes": {n: dag.nodes[n]["conn"] for n in dag.nodes()},
                "connections": build_info["connections"],
            },
            default_flow_style=False,
            stream=fd,
        )


@main.command("run")
@click.option(
    "-s",
    "--select",
    multiple=True,
    default=(),
)
@click.option(
    "-v",
    "--var",
    multiple=True,
)
@click.option("--dry-run/--no-dry-run", "-d/-D", default=False)
def run(select, dry_run, var):

    with open("build/graph.yml") as fd:
        graph = yaml.safe_load(fd)

    connections = {}
    for name, info in graph["connections"].items():
        connections[name] = build_connection(info)

    dag = nx.DiGraph()
    for node, conn in graph["nodes"].items():
        dag.add_node(node, conn=conn)
    for a, b in graph["edges"]:
        dag.add_edge(a, b)

    selected = set(graph["nodes"].keys())
    if select:
        selected = set()
        and_descendants = False
        and_ancestors = False
        for s in select:
            if s.endswith("+"):
                s = s[:-1]
                and_descendants = True
            if s.startswith("+"):
                s = s[1:]
                and_ancestors = True

            selected.add(s)
            if and_descendants:
                selected |= set(nx.descendants(dag, s))
            if and_ancestors:
                selected |= set(nx.ancestors(dag, s))

    for node in nx.topological_sort(dag):
        conn_name = dag.nodes[node]["conn"]
        if node not in selected:
            print(f"Skip {node} (not selected)")
            continue
        conn = connections[conn_name]
        print(f"Execute {node} using connection {conn_name}")
        with open(f"build/{node}.sql") as fd:
            query = fd.read()
            if not dry_run:
                conn.execute(query)


if __name__ == "__main__":
    main()



import os
import sys
from werkzeug.utils import import_string
from jinja2 import Environment


def is_macro(tname: str) -> bool:
    exts = [".jinja2", ".j2", ".sql", ".py"]
    ext = os.path.splitext(tname)[1]
    if ext not in exts:
        return False
    if tname.startswith("macros/"):
        return True
    return False


def load_macros_from_dir(env: Environment):
    macros = {}
    print()
    sys.path.insert(0, "./flow/macros")
    for tname in env.list_templates(filter_func=is_macro):
        template = env.get_template(tname)
        assert template.filename is not None
        if template.filename.endswith(".py"):
            name = os.path.basename(template.filename)[:-3]
            mod = __import__(name, fromlist=["*"])
            all_names = getattr(mod, "__all__", None) or dir(mod)
            public_names = [name for name in all_names if not name.startswith("_")]
            macros[name] = {attr: getattr(mod, attr) for attr in public_names}

        module = template.make_module()
        for name, value in module.__dict__.items():
            if callable(value) and not name.startswith("_"):
                macros[name] = value
    sys.path.pop(0)
    return macros


def load_macros(env: Environment, dag, config):
    def source(name):
        s = config["sources"][name]
        return env.from_string(s).render()

    def env_var(name, *default):
        if not default and name not in os.environ:
            raise ValueError(f"Environment variable {name} is not set")
        return os.environ.get(name, *default)

    def ref(other):
        dag.add_edge(other, env.globals["current_node"])
        return dag.nodes[other].get("ref_name", other)

    def resource(ref_name):
        current_name = env.globals["current_node"]
        dag.add_node(current_name, ref_name=ref_name)
        return ref_name

    def create_table():
        current_name = env.globals["current_node"]
        ref_name = f'"main"."{current_name}"'
        dag.add_node(current_name, ref_name=ref_name)
        return f"""
    DROP TABLE IF EXISTS {ref_name};
    CREATE TABLE {ref_name} AS
    """

    return dict(
        source=source,
        env_var=env_var,
        ref=ref,
        create_table=create_table,
        resource=resource,
    )


def load_macros_from_config(config):
    macros = {}
    sys.path.append(".")
    for item in config.get("macros", []):
        name = item.split(":", 1)[-1]
        macro = import_string(item)
        macros[name] = macro
    sys.path.pop()

    return macros
